"""Check generated HTML."""

import argparse
from collections import defaultdict
from pathlib import Path
import sys

from bs4 import BeautifulSoup
from html5validator.validator import Validator


def main(opt):
    """Main driver."""
    filepaths = Path(opt.dst).glob("**/*.html")
    pages = {path: BeautifulSoup(path.read_text(), "html.parser") for path in filepaths}
    for func in [
        _do_compare_template_readme,
        _do_exercise_titles,
        _do_glossary_redefinitions,
        _do_single_h1,
        lambda o, p: _do_special_links(o, p, "bibliography"),
        lambda o, p: _do_special_links(o, p, "glossary"),
    ]:
        func(opt, pages)
    if opt.html:
        _do_html_validation(opt, pages)


def construct_parser(parser):
    """Parse command-line arguments."""
    parser.add_argument("--dst", type=Path, default="docs", help="output directory")
    parser.add_argument("--html", action="store_true", help="validate HTML")


def _do_compare_template_readme(opt, pages):
    """Compare tables of contents in template and README.md"""
    path = opt.dst / "index.html"
    readme = pages[path]
    for kind, nav_selector, body_selector in [
        ("lessons", "span#nav-lessons", "div#syllabus"),
        ("extras", "span#nav-extras", "div#appendices"),
    ]:
        nav = readme.select(nav_selector)
        if not _require(len(nav) == 1, f"{path} missing or multiple {nav_selector}"):
            continue
        nav = nav[0]
        body = readme.select(body_selector)
        if not _require(len(body) == 1, f"{path} missing or multiple {body_selector}"):
            continue
        body = body[0]

        nav_paths = {node["href"] for node in nav.select("a[href]")}
        body_paths = {node["href"] for node in body.select("a[href]")}
        difference = nav_paths ^ body_paths
        _require(
            not difference,
            f"mis-match in README and nav paths: {', '.join(sorted(difference))}",
        )


def _do_exercise_titles(opt, pages):
    """Check that exercise sections have one <h3> heading."""
    for path, doc in pages.items():
        for section in doc.select("section.exercise"):
            headings = section.select("h1, h2, h3, h4, h5, h6")
            _require(
                len(headings) == 1,
                f"exercise in {path} has missing/multiple heading(s)",
            )
            if len(headings) == 1:
                _require(
                    headings[0].name.lower() == "h3",
                    f"exercise in {path} has {headings[0].name} instead of h3",
                )


def _do_glossary_redefinitions(opt, pages):
    """Check for glossary terms that are defined more than once."""
    seen = defaultdict(list)
    for path, doc in pages.items():
        for node in doc.select("a[href]"):
            if "/glossary/#" in node["href"]:
                key = node["href"].split("#")[-1]
                seen[key].append(path)
    for key, values in seen.items():
        _require(
            len(values) == 1,
            f"glossary entry '{key}' defined in {', '.join(sorted(str(v) for v in values))}",
        )


def _do_html_validation(opt, pages):
    """Validate generated HTML."""
    validator = Validator()
    errors = validator.validate(list(pages.keys()))


def _do_single_h1(opt, pages):
    """Check that each page has exactly one <h1>."""
    for path, doc in pages.items():
        titles = doc.find_all("h1")
        _require(len(titles) == 1, f"{len(titles)} H1 elements found in {path}")


def _do_special_links(opt, pages, stem):
    """Check specially-formatted links."""
    source = opt.dst / stem / "index.html"
    if not _require(source in pages, f"{source} not found"):
        return

    main = pages[source].select("main")
    if not _require(len(main) == 1, f"missing or multiple <main> in {source}"):
        return

    main = main[0]
    defined = {node["id"] for node in main.select("span") if node.has_attr("id")}

    base = f"{stem}/#"
    used = set()
    for path, doc in pages.items():
        here = {
            node["href"].split("#")[-1]
            for node in doc.select("a[href]")
            if base in node["href"]
        }
        used |= here
        unknown = here - defined
        _require(
            len(unknown) == 0,
            f"unknown {stem} keys in {path}: {', '.join(sorted(unknown))}",
        )

    unused = defined - used
    _require(len(unused) == 0, f"unused {stem} keys: {', '.join(sorted(unused))}")


def _require(cond, msg):
    """Check and report."""
    if not cond:
        print(msg, file=sys.stderr)
    return cond



"""Test that sites build."""

import argparse

from bs4 import BeautifulSoup
import pytest

from mccole.build import build, construct_parser, _load_glossary


def test_build_construct_parser_with_default_values():
    parser = argparse.ArgumentParser()
    construct_parser(parser)
    opt = parser.parse_args([])
    assert all(hasattr(opt, key) for key in ["config", "dst", "src", "templates"])


def test_build_with_no_files_creates_empty_output_directory(bare_fs, build_opt):
    build(build_opt)
    dst = bare_fs / build_opt.dst
    assert dst.is_dir()
    assert len(list(dst.iterdir())) == 0


def test_build_with_single_plain_markdown_file_creates_one_output_file(
    bare_fs, build_opt
):
    (bare_fs / build_opt.src / "test.md").write_text("# Title\nbody")
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = BeautifulSoup(expected.read_text(), "html.parser")
    assert doc.title.string == "Title"
    paragraphs = doc.find_all("p")
    assert len(paragraphs) == 1 and paragraphs[0].string == "body"


def test_build_does_not_copy_dot_files(bare_fs, build_opt):
    (bare_fs / build_opt.src / ".gitignore").write_text("content")
    build(build_opt)
    assert len(list((bare_fs / build_opt.dst).iterdir())) == 0


def test_build_does_not_copy_dot_dirs(bare_fs, build_opt):
    (bare_fs / build_opt.src / ".settings").mkdir()
    build(build_opt)
    assert len(list((bare_fs / build_opt.dst).iterdir())) == 0


def test_build_does_not_copy_symlinks(bare_fs, build_opt):
    (bare_fs / build_opt.src / "link.lnk").symlink_to("/tmp")
    build(build_opt)
    assert len(list((bare_fs / build_opt.dst).iterdir())) == 0


def test_build_does_not_copy_destination_files(bare_fs, build_opt):
    (bare_fs / build_opt.dst).mkdir()
    (bare_fs / build_opt.dst / "existing.html").write_text("<html></html>")
    build(build_opt)
    assert len(list((bare_fs / build_opt.dst).iterdir())) == 1


def test_build_does_not_copy_explicitly_skipped_files(bare_fs, build_opt):
    config_file = bare_fs / build_opt.config
    config_file.write_text(
        '[tool.mccole]\nskips = ["*.text", "extras/**", "uv.lock"]\n'
    )
    (bare_fs / build_opt.src / "alpha.text").write_text("something")
    (bare_fs / build_opt.src / "uv.lock").write_text("version = 1")
    (bare_fs / build_opt.src / "extras").mkdir()
    (bare_fs / build_opt.src / "extras" / "test.md").write_text("# test")

    build(build_opt)

    assert len(list((bare_fs / build_opt.dst).iterdir())) == 0


def test_build_boilerplate_files_correctly_renamed(bare_fs, build_opt):
    fixtures = (
        ("CODE_OF_CONDUCT.md", "Code of Conduct", "conduct"),
        ("CONTRIBUTING.md", "Contributing", "contrib"),
        ("LICENSE.md", "License", "license"),
        ("README.md", "Project", ""),
    )
    for filename, content, _ in fixtures:
        x = bare_fs / build_opt.src / filename
        x.write_text(f"# {content}\n")

    build(build_opt)

    for filename, content, output in fixtures:
        expected = bare_fs / build_opt.dst / output / "index.html"
        assert expected.is_file()
        assert content in expected.read_text()


def test_build_boilerplate_links_correctly_adjusted(bare_fs, build_opt):
    lines = (
        "# Title",
        '<section id="text" markdown="1">',
        "[conduct](./CODE_OF_CONDUCT.md)",
        "[contributing](./CONTRIBUTING.md)",
        "[license](./LICENSE.md)",
        "[home page](./README.md)",
        "</section>",
    )
    (bare_fs / build_opt.src / "test.md").write_text("\n".join(lines))

    build(build_opt)

    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = BeautifulSoup(expected.read_text(), "html.parser")
    section = doc.select("section[id='text']")[0]
    urls = {node["href"] for node in section.select("a")}
    assert urls == {"./conduct/", "./contrib/", "./license/", "./"}


def test_build_bibliography_links_correctly_adjusted(bare_fs, build_opt):
    (bare_fs / build_opt.src / "test.md").write_text("# Title\n[](b:first)\n")
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    assert 'href="./bibliography/#first"' in expected.read_text()


def test_build_glossary_links_correctly_adjusted(bare_fs, build_opt):
    (bare_fs / build_opt.src / "test.md").write_text("# Title\n[term](g:key)\n")
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    assert 'href="./glossary/#key"' in expected.read_text()


def test_build_glossary_keys_values_loaded(bare_fs, build_opt, glossary_path):
    glossary = _load_glossary([glossary_path])
    assert glossary == {"first": "first term", "second": "second term"}


def test_build_defined_terms_added_to_page(bare_fs, build_opt, glossary_path):
    lines = ["# Title", '<p id="terms"></p>', "[one](g:first) [two](g:second)"]
    (bare_fs / build_opt.src / "test.md").write_text("\n".join(lines))
    build(build_opt)
    content = (bare_fs / build_opt.dst / "test.html").read_text()
    doc = BeautifulSoup(content, "html.parser")
    terms = doc.select("p#terms")[0]
    refs = {node["href"] for node in terms.select("a[href]")}
    assert refs == {"./glossary/#first", "./glossary/#second"}


def test_build_defined_terms_paragraph_removed(bare_fs, build_opt):
    lines = ["# Title", '<p id="terms"></p>', "body"]
    (bare_fs / build_opt.src / "test.md").write_text("\n".join(lines))
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = BeautifulSoup(expected.read_text(), "html.parser")
    assert not doc.select("p#terms")


def test_build_backtick_code_block_class_applied_to_enclosing_pre(bare_fs, build_opt):
    (bare_fs / build_opt.src / "test.md").write_text("# Title\n```py\nx = 1\n```\n")
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = BeautifulSoup(expected.read_text(), "html.parser")
    for tag in ["pre", "code"]:
        nodes = doc.find_all(tag)
        assert len(nodes) == 1
        assert nodes[0]["class"] == ["language-py"]


def test_build_non_markdown_files_copied(bare_fs, build_opt):
    (bare_fs / build_opt.src / "in_root.txt").write_text("root text")
    (bare_fs / build_opt.src / "subdir").mkdir()
    (bare_fs / build_opt.src / "subdir" / "in_subdir.txt").write_text("subdir text")
    build(build_opt)
    in_root = bare_fs / build_opt.dst / "in_root.txt"
    assert in_root.is_file()
    assert in_root.read_text() == "root text"
    in_subdir = bare_fs / build_opt.dst / "subdir" / "in_subdir.txt"
    assert in_subdir.is_file()
    assert in_subdir.read_text() == "subdir text"


def test_build_warn_unknown_markdown_links(bare_fs, build_opt, capsys):
    (bare_fs / build_opt.src / "test.md").write_text("# Title\n[text](link.md)\n")
    build(build_opt)
    captured = capsys.readouterr()
    assert "unknown Markdown link" in captured.err


def test_build_warn_missing_h1(bare_fs, build_opt, capsys):
    (bare_fs / build_opt.src / "test.md").write_text("text\n")
    build(build_opt)
    captured = capsys.readouterr()
    assert "lacks H1 heading" in captured.err


def test_build_warn_missing_title(bare_fs, build_opt, capsys):
    template_path = bare_fs / build_opt.templates / "page.html"
    template = template_path.read_text()
    template = template.replace("<title></title>", "")
    template_path.write_text(template)

    (bare_fs / build_opt.src / "test.md").write_text("text\n")
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not have <title> element" in captured.err


def test_build_warn_badly_formatted_config_file(bare_fs, build_opt, capsys):
    config_file = bare_fs / "pyproject.toml"
    config_file.write_text('[tool.missing]\nskips = ["extras", "uv.lock"]\n')

    build(build_opt)
    captured = capsys.readouterr()
    assert "does not have 'tool.mccole'" in captured.err


def test_build_warn_overlap_renames_and_skips(bare_fs, build_opt, capsys):
    config_file = bare_fs / "pyproject.toml"
    config_file.write_text('[tool.mccole]\nskips = ["LICENSE.md"]\n')
    (bare_fs / build_opt.src / "LICENSE.md").write_text("# License")

    build(build_opt)
    captured = capsys.readouterr()
    assert "overlap between skips and renames" in captured.err


def test_build_warn_multiple_term_paragraphs_in_doc(bare_fs, build_opt, capsys):
    lines = [
        "# Terms",
        '<p id="terms"></p>',
        '<p id="terms"></p>',
    ]
    (bare_fs / build_opt.src / "test.md").write_text("\n".join(lines))

    build(build_opt)
    captured = capsys.readouterr()
    assert "terms paragraph appears multiple times" in captured.err


def test_build_warn_multiple_glossary_files_found(bare_fs, build_opt, capsys):
    (bare_fs / build_opt.src / "glossary").mkdir()
    (bare_fs / build_opt.src / "glossary" / "index.md").write_text("# First")
    (bare_fs / build_opt.src / "subdir").mkdir()
    (bare_fs / build_opt.src / "subdir" / "glossary").mkdir()
    (bare_fs / build_opt.src / "subdir" / "glossary" / "index.md").write_text(
        "# Second"
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "multiple glossary files" in captured.err



"""Test site checker."""

import argparse
import re

from bs4 import BeautifulSoup
import pytest

from mccole.build import build, construct_parser as build_parser
from mccole.lint import lint, construct_parser as lint_parser

DIV_APPENDICES = '<div id="appendices"></div>'
NAV_LESSONS = '<span class="dropdown-content" id="nav-lessons"></span>'

README = """\
# README
<div id="syllabus"></div>
<div id="appendices"></div>
"""


@pytest.fixture
def lint_fs(build_opt, bare_fs):
    (bare_fs / "README.md").write_text(README)
    (bare_fs / "bibliography").mkdir()
    (bare_fs / "bibliography" / "index.md").write_text("# Bibliography")
    (bare_fs / "glossary").mkdir()
    (bare_fs / "glossary" / "index.md").write_text("# Glossary")
    return bare_fs


def test_lint_construct_parser_with_default_values():
    parser = argparse.ArgumentParser()
    lint_parser(parser)
    opt = parser.parse_args([])
    assert hasattr(opt, "dst")


def test_lint_no_problems_to_report(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert not captured.err


def test_lint_multiple_h1_in_file(build_opt, lint_opt, lint_fs, capsys):
    (lint_fs / "test.md").write_text("# First\n\n# Second\n")
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    m = re.search(r"(\d) H1 elements found", captured.err)
    assert m and int(m.group(1)) == 2


@pytest.mark.parametrize("kind", ["bibliography", "glossary"])
def test_special_file_not_found(build_opt, lint_opt, lint_fs, capsys, kind):
    build(build_opt)
    (lint_fs / lint_opt.dst / kind / "index.html").unlink()
    lint(lint_opt)
    captured = capsys.readouterr()
    assert re.search(f"{kind}/index.html not found", captured.err)


@pytest.mark.parametrize("kind,key", [["bibliography", "b:key"], ["glossary", "g:key"]])
def test_special_key_undefined(build_opt, lint_opt, lint_fs, capsys, kind, key):
    lines = [
        "# A",
        f"[]({key})",
    ]
    (lint_fs / "a.md").write_text("\n".join(lines))

    build(build_opt)
    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search(f"unknown {kind} keys in .+: key", captured.err)


@pytest.mark.parametrize("kind", ["bibliography", "glossary"])
def test_special_key_unused(build_opt, lint_opt, lint_fs, capsys, kind):
    build(build_opt)

    lines = [
        "# Title",
        "<main>",
        '<span id="key">text</span>',
        "</main>",
    ]
    (lint_fs / lint_opt.dst / kind / "index.html").write_text("\n".join(lines))

    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search(f"unused {kind} keys: key", captured.err)


def test_multiple_main_in_special_file(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    page = build_opt.dst / "bibliography" / "index.html"
    content = page.read_text()
    content = content.replace("<main>", "<main></main><main>")
    page.write_text(content)

    lint(lint_opt)
    captured = capsys.readouterr()
    assert "missing or multiple" in captured.err


def test_glossary_term_redefined(build_opt, lint_opt, lint_fs, capsys):
    a = [
        "# A",
        "[term 1](g:term_1)",
        "[term 1 again](g:term_1)",
        "[term 2](g:term_2)",
    ]
    (lint_fs / "a.md").write_text("\n".join(a))

    b = [
        "# B",
        "[term 2](g:term_2)",
    ]
    (lint_fs / "b.md").write_text("\n".join(b))

    g = [
        "# Glossary",
        '<span id="term_1">term 1</span>',
        '<span id="term_2">term 2</span>',
    ]
    (lint_fs / "glossary" / "index.md").write_text("\n".join(g))

    build(build_opt)
    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search(r"glossary entry 'term_1' defined in .+, .+", captured.err)
    assert re.search(r"glossary entry 'term_2' defined in .+, .+", captured.err)


def test_glossary_key_unused(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)

    g = [
        "# Glossary",
        "<main>",
        '<span id="key">text</span>',
        "</main>",
    ]
    (lint_fs / lint_opt.dst / "glossary" / "index.html").write_text("\n".join(g))

    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search("unused glossary keys: key", captured.err)


@pytest.mark.parametrize(
    "headings,msg",
    [
        ["", "has missing/multiple heading(s)"],
        ["## first\n## second", "has missing/multiple heading(s)"],
        ["# level 1", "has h1 instead of h3"],
        ["## level 2", "has h2 instead of h3"],
        ["#### level 4", "has h4 instead of h3"],
    ],
)
def test_exercise_section_has_bad_headings(
    build_opt, lint_opt, lint_fs, capsys, headings, msg
):
    lines = [
        "# Title",
        '<section class="exercise" markdown="1">',
        headings,
        "</section>",
    ]
    (lint_fs / build_opt.src / "test.md").write_text("\n".join(lines))

    build(build_opt)
    lint(lint_opt)

    captured = capsys.readouterr()
    assert msg in captured.err


def test_html_validation_with_valid_html(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    lint_opt.html = True
    lint(lint_opt)
    captured = capsys.readouterr()
    assert not captured.err


def test_html_validation_with_invalid_html(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    page = build_opt.dst / "index.html"
    content = page.read_text()
    page.write_text(content.replace('<html lang="en">', ""))

    lint_opt.html = True
    lint(lint_opt)
    captured = capsys.readouterr()
    assert captured.out


@pytest.mark.parametrize(
    "target,replacement",
    [
        [NAV_LESSONS, ""],
        [NAV_LESSONS, f"{NAV_LESSONS}\n{NAV_LESSONS}"],
        [DIV_APPENDICES, ""],
        [DIV_APPENDICES, f"{DIV_APPENDICES}\n{DIV_APPENDICES}"],
    ],
)
def test_compare_template_readme_missing_nav(
    build_opt, lint_opt, lint_fs, capsys, target, replacement
):
    build(build_opt)
    page = build_opt.dst / "index.html"
    content = page.read_text()
    content = content.replace(target, replacement)
    page.write_text(content)

    lint_opt.html = True
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "missing or multiple" in captured.err
